/* 
  Particle.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  27-SEP-1996

*/

#import <math.h>
#import <collections.h>
#import <simtools.h>

#import "Lattice.h"
#import "Bernoulli.h"
#import "Particle.h"

@implementation Particle


static double defaultMobilityFactor = 1.0;
static id bernoulli;
static id uniformUnsigned;
static BOOL energeticMotionFlag;

/*
  These are sortof, kindof, class variables (as opposed to
  instance variables). They should only be accessed
  through the relevant +set and +get *class* methods. If 
  subclasses want have something
  other than the default values they will have to
  provide their own class variables *and* override the
  relevant +set and +get methods...
*/

+(void) setClassMobilityFactor: (double) inMobilityFactor {
  defaultMobilityFactor = inMobilityFactor;
}

+(double) getClassMobilityFactor {
  return(defaultMobilityFactor);
}

+(void) setBernoulli: (id) inBernoulli {
  bernoulli = inBernoulli;
}

+(id) getBernoulli {
  return(bernoulli);
}

+(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
}

+(id) getUniformUnsigned {
  return(uniformUnsigned);
}

+(void) setEnergeticMotionFlag: (BOOL) value {
  energeticMotionFlag = value;
}

+(BOOL) getEnergeticMotionFlag {
  return(energeticMotionFlag);
}

-(double) getMobilityFactor {
  // This is really a hook-method: subclasses may
  // override it to allow instance objects to provide
  // an instance specific mobility factor.  However,
  // the default behavior, defined here in the default
  // method, is simply to provide the class-wide
  // mobility factor.
  return [[self class] getClassMobilityFactor];
}

-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode {
  /* 
    neighborCode is provided for generality of the interface -
    in the general case, the ability to move may depend on
    the direction proposed. But in the present (default)
    case, it is ignored...  */
  return (!hasMoved);
}

-(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode { 
  /* In the simple, default, case, ability to move is
     symettrical for both particles (primary and
     secondary); but in general they may have different
     views...  */
  return [self canMoveToNeighborCode: neighborCode];
}


-(void) prepareForMotion {
  //  This is just a hook, so agents can - if they wish
  //  - remember something prior to a movement...

}


-setHasMoved {
  hasMoved = YES;

  return self;
}

-initialiseStep {
  hasMoved = NO;

  return self;
}

-doEnergeticMotion {

  /* NB: It's important to use [self getMobilityFactor] here
     (rather than referring directly to defaultMobilityFactor)
     because we may be executing -doMotion in a subclass with
     a different mobilityFactor. */

  NeighborCode neighborCode, initialNeighborCode;
  Particle *neighborParticle;
  double motionProbability;
  id eligibleNeighborParticles;
  unsigned count, randOffset;
  id secondaryParticle;

  secondaryParticle = nil;
  eligibleNeighborParticles = [List create: [self getZone]];
  initialNeighborCode = [[coord getLatticeObject] getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    if ([self canMoveToNeighborCode: neighborCode]) {
      neighborParticle = [self getAgentAtNeighborCode: neighborCode];

      if (neighborParticle == nil) 
	[InternalError raiseEvent:
	   "nil encountered at neighborCode %d of particle %d.\n",
	   neighborCode, self];
      else {
	if ([neighborParticle canBeMovedToNeighborCode: 
	    [[coord getLatticeObject] 
	      getOppositeNeighborCode: neighborCode]]) {
	  [eligibleNeighborParticles addLast: neighborParticle];
	}
      }
    }
    neighborCode = 
      [[coord getLatticeObject] getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);

  count = [eligibleNeighborParticles getCount];
  if (count > 0) {
    if (count == 1)
      randOffset = 0;
    else
      randOffset = [uniformUnsigned getUnsignedWithMin: 0
		    withMax: (count - 1)];
    neighborParticle = [eligibleNeighborParticles atOffset: randOffset];
    [self prepareForMotion];
    hasMoved = [self attemptSwapWithAgent: neighborParticle];
    if (hasMoved)  {
    //    motionProbability = [self getMobilityFactor];
      motionProbability = sqrt([self getMobilityFactor] * 
 	[neighborParticle getMobilityFactor]);
      if ([bernoulli getTossWithProb: motionProbability]) {
	[neighborParticle setHasMoved];
	secondaryParticle = neighborParticle;
      }
      else { // Backout!
	if (![self attemptSwapWithAgent: neighborParticle]) 
	  [InternalError raiseEvent:
	     "Couldn't back out of motion???\n"];
      }
    }
  }

  [eligibleNeighborParticles drop];
  return secondaryParticle;
}

-doLazyMotion {

  /* NB: It's important to use [self getMobilityFactor] here
     (rather than referring directly to defaultMobilityFactor)
     because we may be executing -doMotion in a subclass with
     a different mobilityFactor. */

  NeighborCode neighborCode;
  Particle *neighborParticle;
  double motionProbability;
  id secondaryParticle;

  secondaryParticle = nil;
  neighborCode = [[coord getLatticeObject] getRandomNeighborCode];

  if ([self canMoveToNeighborCode: neighborCode]) {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle canBeMovedToNeighborCode: 
          [[coord getLatticeObject] getOppositeNeighborCode: neighborCode]]) {
        [self prepareForMotion];
	hasMoved = [self attemptSwapWithAgent: neighborParticle];
	if (hasMoved)  {
	  //  motionProbability = [self getMobilityFactor];
	  motionProbability = sqrt([self getMobilityFactor] * 
	    [neighborParticle getMobilityFactor]);
	  if ([bernoulli getTossWithProb: motionProbability]) {
	    [neighborParticle setHasMoved];
	    secondaryParticle = neighborParticle;
	  }
	  else { // Backout!
	    if (![self attemptSwapWithAgent: neighborParticle]) 
	      [InternalError raiseEvent:
		 "Couldn't back out of motion???\n"];

            [neighborParticle setHasMoved];
	    secondaryParticle = neighborParticle;
	  }
	}
      }
    }
  }

  return secondaryParticle;
}

-(id) doMotion {
  id secondaryParticle;

  if (energeticMotionFlag)
    secondaryParticle = [self doEnergeticMotion];
  else
    secondaryParticle = [self doLazyMotion];

  return secondaryParticle;
}

-step {
  [self doMotion]; // Only default behaviour is doMotion.

  return self;
}


@end
