/* 
  SCLDisplay.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  MAR-1997

*/

#import <simtools.h>
#import <simtoolsgui.h>

#import "LatticeAgent.h"
#import "SCLGlobals.h"
#import "Lattice2d.h"
#import "Link.h"
#import "Bond.h"
#import "ParticleManager.h"
#import "OrientationCode.h"
#import "Coord2d.h"
#import "UserCellEditor.h"
#import "SCLDisplay.h"

@implementation SCLDisplay

typedef enum {
  WHITE, RED, LIMEGREEN, BLUE, SKYBLUE, STEELBLUE, CYAN, BLACK}
  colors;

static const int rasterScaleFactor = 8;

-createEnd {
  SCLDisplay *finalSelf;
  id coord;

  finalSelf = [super createEnd];

  if (finalSelf->lattice == nil)
    [InternalError raiseEvent:
       "SCLDisplay object *must* be initialised with a lattice.\n"];

  finalSelf->colormap = [Colormap create: [finalSelf getZone]];
  [finalSelf->colormap setColor: WHITE     ToName: "white"];
  [finalSelf->colormap setColor: RED       ToName: "red"];
  [finalSelf->colormap setColor: LIMEGREEN ToName: "LimeGreen"];
  [finalSelf->colormap setColor: BLUE      ToName: "blue"];
  [finalSelf->colormap setColor: SKYBLUE   ToName: "SkyBlue"];
  [finalSelf->colormap setColor: STEELBLUE ToName: "SteelBlue"];
  [finalSelf->colormap setColor: CYAN      ToName: "cyan"];
  [finalSelf->colormap setColor: BLACK     ToName: "black"];
  
  finalSelf->backgroundColor = WHITE;

  [finalSelf->lattice setRasterScaleFactor: rasterScaleFactor];
  finalSelf->raster = [ZoomRaster create: [self getZone]];
  [finalSelf->raster setColormap: colormap];
  [finalSelf->raster setZoomFactor: 2];

  coord = [lattice createMaxCoord: [finalSelf getZone]];
  [finalSelf->raster setWidth: ([coord getRasterX] + rasterScaleFactor) 
                  Height: ([coord getRasterY] + rasterScaleFactor) ];
  [coord drop];

  [finalSelf->raster setWindowTitle: "SCL Lattice"];
  [finalSelf->raster pack];

  [finalSelf->raster setButton: ButtonRight Client: finalSelf 
    Message: M(makeProbeAtX:Y:)];
  [finalSelf->raster setButton: ButtonLeft Client: finalSelf 
    Message: M(setCellEditorX:Y:)];

  return (finalSelf);
}

-(void) setCellEditorX: (int) rasterX Y: (int) rasterY {
  // This is just a stub, passing the message on to
  // userCellEditor. I do things in this clumsy way,
  // rather than programming the "direct" message into
  // the raster, because the userCellEditor may not yet
  // have been created at the time I am configuring the
  // raster...
  [userCellEditor setRasterX: rasterX Y: rasterY];
}


-(void) drop {
  [raster drop];
  [colormap drop];

  [super drop];
}

-(void) setBackgroundColor: (Color) color {
  backgroundColor = color;
}

-(Color) getBackgroundColor {
  return backgroundColor;
}

-(void) setLatticeToDisplay: (id) inLattice {
  lattice = inLattice;
}

-(void) eraseRaster {
  [raster fillRectangleX0: 0 Y0: 0
    X1: [raster getWidth] Y1: [raster getHeight] 
    Color: backgroundColor];
}

-(void) displaySubstrates {
  id substrateList, index;
  id substrate;
  id coord;
  int rasterX, rasterY;

  substrateList = [substrateMgr getAgentList];
  index = [substrateList begin: [self getZone]];
  while ((substrate = [index next]) != nil) {
    coord = [substrate copyCoord: [self getZone]];
    rasterX = [coord getRasterX];
    rasterY = [coord getRasterY];
    [coord drop];

    rasterX++;
    rasterY++;

    [raster drawPointX: rasterX   Y: rasterY++ Color: RED];
    [raster drawPointX: rasterX   Y: rasterY++ Color: RED];
    [raster drawPointX: rasterX++ Y: rasterY   Color: RED];
    [raster drawPointX: rasterX++ Y: rasterY   Color: RED];
    [raster drawPointX: rasterX   Y: rasterY-- Color: RED];
    [raster drawPointX: rasterX   Y: rasterY-- Color: RED];
    [raster drawPointX: rasterX-- Y: rasterY   Color: RED];
    [raster drawPointX: rasterX   Y: rasterY   Color: RED];
  }
  [index drop];
}

-(void) displayLinks {
  id linkList, index;
  id link;
  id coord;
  int rasterX, rasterY, baseRasterX, baseRasterY;
  Color color;

  linkList = [linkMgr getAgentList];
  index = [linkList begin: [self getZone]];
  while ((link = [index next]) != nil) {
    coord = [link copyCoord: [self getZone]];
    baseRasterX = [coord getRasterX];
    baseRasterY = [coord getRasterY];
    [coord drop];

    if ([link isDisintegrating]) color = STEELBLUE;
      else 
        if ([link isPromoting])
          color = SKYBLUE;
        else
          color = BLUE;

    rasterX = baseRasterX;
    rasterY = baseRasterY;
    [raster drawPointX: rasterX   Y: rasterY++ Color: color];
    [raster drawPointX: rasterX   Y: rasterY++ Color: color];
    [raster drawPointX: rasterX   Y: rasterY++ Color: color];
    [raster drawPointX: rasterX   Y: rasterY++ Color: color];
    [raster drawPointX: rasterX++ Y: rasterY   Color: color];
    [raster drawPointX: rasterX++ Y: rasterY   Color: color];
    [raster drawPointX: rasterX++ Y: rasterY   Color: color];
    [raster drawPointX: rasterX++ Y: rasterY   Color: color];
    [raster drawPointX: rasterX   Y: rasterY-- Color: color];
    [raster drawPointX: rasterX   Y: rasterY-- Color: color];
    [raster drawPointX: rasterX   Y: rasterY-- Color: color];
    [raster drawPointX: rasterX   Y: rasterY-- Color: color];
    [raster drawPointX: rasterX-- Y: rasterY   Color: color];
    [raster drawPointX: rasterX-- Y: rasterY   Color: color];
    [raster drawPointX: rasterX-- Y: rasterY   Color: color];
    [raster drawPointX: rasterX   Y: rasterY   Color: color];


    if ([link hasAbsorbedSubstrate]) {
      rasterX = baseRasterX + 1;
      rasterY = baseRasterY + 1;

      [raster drawPointX: rasterX   Y: rasterY++ Color: RED];
      [raster drawPointX: rasterX   Y: rasterY++ Color: RED];
      [raster drawPointX: rasterX++ Y: rasterY   Color: RED];
      [raster drawPointX: rasterX++ Y: rasterY   Color: RED];
      [raster drawPointX: rasterX   Y: rasterY-- Color: RED];
      [raster drawPointX: rasterX   Y: rasterY-- Color: RED];
      [raster drawPointX: rasterX-- Y: rasterY   Color: RED];
      [raster drawPointX: rasterX   Y: rasterY   Color: RED];
    }
  }
  [index drop];
}

-(void) displayCatalysts {
  id catalystList, index;
  id catalyst;
  id coord;
  int rasterX, rasterY;
  int i, j;

  catalystList = [catalystMgr getAgentList];
  index = [catalystList begin:[ self getZone]];
  while ((catalyst = [index next]) != nil) {
    coord = [catalyst copyCoord: [self getZone]];
    rasterX = [coord getRasterX];
    rasterY = [coord getRasterY];
    [coord drop];

    rasterY++;
    for (j = 0; j < 3; j++) {
      [raster drawPointX: rasterX Y: rasterY Color: LIMEGREEN];
      rasterY++;
    }
    rasterY -= 4;
    rasterX++;

    for (i = 0; i < 3; i++) {
      for (j = 0; j < 5; j++) {
	[raster drawPointX: rasterX Y: rasterY Color: LIMEGREEN];
	rasterY++;
      }
    rasterY -= 5;
    rasterX++;
    }

    rasterY++;
    for (j = 0; j < 3; j++) {
      [raster drawPointX: rasterX Y: rasterY Color: LIMEGREEN];
      rasterY++;
    }
  }
  [index drop];
}


-(void) displayBonds {
  id bondList, index;
  id bond, link0, link1;
  id coord;
  int rasterX, rasterY;
  int x0, x1, y0, y1;
  int j;
  OrientationCode orientationCode;
  BOOL oddColumn;

  bondList = [bondMgr getAgentList];
  index = [bondList begin: [self getZone]];
  while ((bond = [index next]) != nil) {
    link0 = [bond getLink0];
    link1 = [bond getLink1];

    if ((link0 == nil) || (link1 == nil))
      [InternalError raiseEvent:
	 "Nil link encountered on bond %d.\n"
	 "link0: %d  link1: %d\n",
	 bond, link0, link1];

    coord = [link0 copyCoord: [self getZone]];
    x0 = [coord getX];
    y0 = [coord getY];
    [coord drop];
    oddColumn = ((x0 % 2) == 1);

    coord = [link1 copyCoord: [self getZone]];
    x1 = [coord getX];
    y1 = [coord getY];
    [coord drop];

    rasterX = (4 * (x0 + x1)) + 1; // Simple case...

    // Now allow for bond at edge of raster...
    if (x0 != x1) {
      if (x0 == 0) {
	if (x1 != 1)
	  rasterX = (rasterScaleFactor * x1) + 5;
      }
      else {
	if (x1 == 0) {
	  if (x0 != 1)
	    rasterX = (rasterScaleFactor * x0) + 5;
	}
      }
    }

    rasterY = (4 * (y0 + y1)) + 1; // Simple case...

    // Now allow for bond at edge of raster...
    if (y0 != y1) {
      if (y0 == 0) {
	if (y1 != 1)
	  rasterY = (rasterScaleFactor * y1) + 5;
      }
      else {
	if (y1 == 0) {
	  if (y0 != 1)
	    rasterY = (rasterScaleFactor * y0) + 5;
	}
      }
    }

    orientationCode = [bond getOrientationCode];

    switch (orientationCode) {
    case TriNorthSouth :
      if (oddColumn)
	rasterY += 4;
    case SqrNorthSouth :
      rasterX++;
      for (j = 0; j < 3; j++) 
	[raster drawPointX: rasterX Y: rasterY++ Color: BLUE];
      break;

    case SqrEastWest:
      rasterY++;
      for (j = 0; j < 3; j++) 
	[raster drawPointX: rasterX++ Y: rasterY Color: BLUE];
      break;

    case TriNorthEastSouthWest:
      rasterY += 2;
    case SqrNorthEastSouthWest:
      rasterX += 2;
      for (j = 0; j < 3; j++) 
	[raster drawPointX: rasterX-- Y: rasterY++ Color: BLUE];
      break;

    case TriNorthWestSouthEast:
      rasterY += 2;
    case SqrNorthWestSouthEast:
      for (j = 0; j < 3; j++) 
	[raster drawPointX: rasterX++ Y: rasterY++ Color: BLUE];
      break;

    default:
      [InternalError raiseEvent:
       "Out-of-range orientationCode (%d) encountered.\n", orientationCode];
    }
  }
  [index drop];
}

-(void) displayLattice {
  [self eraseRaster];

  // Don't display holes, because they just *are* the
  // same as the background...
  [self displaySubstrates];
  [self displayLinks];
  [self displayBonds];
  [self displayCatalysts];

  [raster drawSelf];
}

-makeProbeAtX: (int) rasterX Y: (int) rasterY {
  id agent;
  id coord;

  coord = [lattice createOriginCoord: [self getZone]];
  [coord setRasterX: rasterX];
  [coord setRasterY: rasterY];

  agent = [coord getObject];
  if (agent != nil)
    [probeDisplayManager createProbeDisplayFor: agent];

  else

  //  [globalTkInterp eval: "bell"];

  // NB: use of globalTkInterp was deprecated from
  // 1.0.4 onward; support was removed completely (?)
  // in 1.1. I dunno how to achieve the bell effect
  // without it...

    [WarningMessage raiseEvent: 
      "makeProbeAtX:Y: no agent present at position..."];

  [coord drop];

  return self;
}
      
@end
